/*
   program : backproject.c

   $Source: /home/yee/C/RCS/parallel.c,v $
   $Author: yee $
   $Date: 1999/02/10 16:39:47 $
   $Revision: 1.17 $

   This file contains the code implementing parallel beam backprojection.
   It is called from code in backproject.c

   Compile Options :
      1) #define RECON_INTERPOLATE for linear interpolation of backprojection
         data or

         #define RECON_NEAREST for nearest neighbor interpolation of data.
  RECON_NEAREST is the default.

      2) #define RECON_DOUBLE|DOUBLE for double precision.  This file defaults
         to #define SINGLE for single precision.

      3) #define RECON_PARALLEL if support for parallel projection data is to
         be supported.

  #define RECON_PARALLEL_FAN if parallel projection data is to be
  reconstructed as a special case of fan beam data. (slower than native
  parallel projection reconstruction)

   Runtime Parameters : none

   Assumptions : none known

   Bugs :
     1) ParallelBackprojectionProjection() no longer works for non-square
        pixels.

   To Do :
     1) clean up documentation over the definition of theta.  Is it the angle
        of the projection or of its perpendicular?

   $Log: parallel.c,v $
   Revision 1.17  1999/02/10 16:39:47  yee
   *** empty log message ***

   Revision 1.16  1999/01/01 04:25:16  yee
   *** empty log message ***

   Revision 1.15  1998/12/21 15:40:17  yee
   working parallel. broken angle/colin

   Revision 1.14  1998/08/27 22:54:29  yee
   working for parallel and colinear fan geometries

   Revision 1.10  1997/11/09 20:59:08  yee
   eliminated DetermineRay().

   Copyright 1997 by Conway Yee.  All rights reserved.  */

char parallel[] = "$Id: parallel.c,v 1.17 1999/02/10 16:39:47 yee Exp $";

#include <math.h>

#include "backproject.h"
#include "codes.h"
#include "style.h"
#include "ImageHeader.h"
#include "filter.h"
 
/*****************************************************************************/
/*****************************************************************************/

/* this function weights the backprojection filter for fan beam data.  It
   is called by the generic filter code in filter.c */

#ifdef RECON_PARALLEL_FAN
PUBLIC double ParallelBackprojectionFilterWeight
(
 int ndim,           /* dimensionality of data array */
 int i,              /* element number */
 void *free          /* degrees of freedom */
)
{ /* ParallelProjectionFilterWeight() */
  return(1.0);
} /* ParallelProjectionFilterWeight() */
#endif RECON_PARALLEL_FAN

/*****************************************************************************/

/* this function weights the projections for fan beam data */

#ifdef RECON_PARALLEL_FAN
PUBLIC double ParallelProjectionWeight
(
 int ndim,           /* dimensionality of data array */
 int i,              /* element number */
 void *free          /* degrees of freedom */
)
{ /* ParallelProjectionWeight() */
  return(1.0);
} /* ParallelProjectionWeight() */
#endif RECON_PARALLEL_FAN

/*****************************************************************************/

/* this function calculates the point from which a pixel is backprojected
   in units of distance */

#ifdef RECON_PARALLEL_FAN
PUBLIC double ParallelBackprojectPoint
(
 double x,          /* X position in 2D image space */
 double y,          /* Y position in 2D image space */
 double *qtau,      /* spacing of projection elements */
 double *ftau,      /* spacing of reconstructed image elements */
 double sin_t,      /* sin of angle made by projection with +X axis */
 double cos_t,      /* cos of angle made by projection with +X axis */
 double D           /* distance between origin and focal point */
)
{ /* ParallelBackprojectPoint() */
                             /* old function when non-square pixels worked */
  /*  return((i*sin_t/ftau[X]-j*cos_t/ftau[Y])*qtau[ELE]*qtau[ELE]); */
  return(x*sin_t-y*cos_t);
} /* ParallelBackprojectPoint() */
#endif RECON_PARALLEL_FAN

/*****************************************************************************/

/* this function is used by BackprojectProjection() for the normalization
   needed for each point in the backprojection.  It is not needed for
   parallel beam data. */

#ifdef RECON_PARALLEL_FAN
PUBLIC double P2
(
 double D,      /* distance between center of rotation and focal point */
 double x,      /* X coordinate of point */
 double y,      /* Y coordinate of point */
 double sin_t,  /* sin(angle of center ray of projection) */
 double cos_t   /* cos(angle of center ray of projection) */
)
{ /* P2() */
  return(1.0);
} /* P2() */
#endif RECON_PARALLEL_FAN

/*****************************************************************************/

/* this function performs back projection on parallel beam data

   This function was not folded into a generic BackprojectProjection()
   function because of the performance penalties involved.  At last
   measurement, Such a function shows an approximately 50% performance
   penalty.

   NOTE: cos(theta) and sin(theta) are sent in as parameters to allow the
         use of recurrence formulae.   */

#ifdef RECON_PARALLEL_NATIVE
PUBLIC void ParallelBackprojectProjection
(
 RECON_FLOAT **f,           /* reconstructed image */
 int *flim,                 /* dimensionality of f[][] */
 double *ftau,              /* dist between samples in reconstructed image */
 RECON_COMPLEX *q,          /* projection data */
 int *qlim,                 /* dimensionality of q[] */
 double *qtau,              /* distance between samples in projection data */
 double theta,              /* angle of projection data */
 double D,                  /* dist between focal pt and center of rotation */
 double cos_t,              /* cos(angle of projection data) */
 double sin_t,              /* sin(angle of projection data) */
 BackprojectFunctions *func /* unused ptr to structure with function choices */
)
{ /* ParallelBackprojectProjection() */
  int i,j;      /* counters */

  q -= qlim[LO_ELE];  /* BackprojectProjection() uses offset information */

  for(i=flim[LOX];i<=flim[HIX];i++) {
    for(j=flim[LOY];j<=flim[HIY];j++) {
      double ProjectedPosition=i*sin_t-j*cos_t; /*projected position of (i,j)*/

#ifdef RECON_INTERPOLATE
      { /* interpolate */
 int lo=(int)floor(ProjectedPosition);  /* low end of interpolation */
 double frac=ProjectedPosition-lo;      /* fraction between pixels */

 if((lo>=qlim[LO_ELE])&&(lo<qlim[HI_ELE]))
   f[i][j] += frac*q[lo].r+(1-frac)*q[lo+1].r;
      } /* interpolate */
#else RECON_NEAREST
      {
 int ele=NINT(ProjectedPosition);   /* element of q used */

 if((ele>=qlim[LO_ELE])&&(ele<=qlim[HI_ELE]))
   f[i][j] += q[ele].r;
      } 
#endif
    }
  }
} /* ParallelBackprojectProjection() */
#endif RECON_PARALLEL_NATIVE
